/*
 * This is FXcube, a tiny Linux demo where I collected my favourite
 * oldskool effects on a single 3d spinning cube! Have fun. To skip
 * scenes, press keys. Some 100 MHz needed. You need X11 with 16bit
 * colors and 1.5 MB of free RAM, plus /dev/dsp for fake music :-).
 * Copyright 1994-2003 by Eric Auer <eric@coli.uni-sb.de>.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should find a copy of the GNU General Public License on the
 * site where you have downloaded this source. If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307 USA
 * (or try http://www.gnu.org/licenses/licenses.html at www.gnu.org).
 */

#include "main.h"


/* limit the cycles for each effect here. Pressing a key in the win */
/* also advances to the next effect. Cube has fixed maximum time.   */
#define MUSTERCOUNT 50  /* muster is only cyclic (70) in all / cube */
#define INTERCOUNT 100
#define LASERCOUNT 100
#define PLASMACOUNT 64
#define APFELCOUNT (4 * abs(INVSTEPS))
#define FIRECOUNT  100
#define STARCOUNT  100
#define ALLCOUNT   140

#define FORK 42
// #undef FORK

#ifdef FORK
#define UPDATE show_x(); (void)gettime(); 
#else
#define UPDATE show_x(); oss_fillbuffer(&musikfueller);
#endif

/* evtl. noch mehr, z.B. testen, ob Audio-Nachschub noetig ist, etc.     */
/* gettime wird z.B. als Lebenszeichen an den geforkten Prozess gerufen! */
/* fillbuffer geht nur mit ioctl gut, aber nur 1 Prozess laeuft - das    */
/* ist leichter im Griff zu behalten!                                    */

double POWi(double x, int p) // Aus include/i386/__math.h abgeleitet
{ double r=1.0;              // normal ist pow(double,double) -> double
  if (p < 0) { p = -p; x = 1.0/x; }
  for(;;) 
  { if (p & 1) r *= x;
    if ((p >>= 1) == 0) return r;
    x *= x;
  };
};

/* 
 * MAIN ist ein define fuer main oder _start (falls -nostartfiles)
 */
int MAIN () 
{ 

char * mustername = "some simple pattern  ";
char * starname =   "3d-Starfield         ";
char * intername =  "Interferenz von 2xSin";
char * lasername =  "Laser-Polar-Lissajous";
char * plasmaname = "Plasma, Virt.Pal.Ani.";
char * apfelname =  "Animierter Apfelmann ";
char * firename =   "really nice FIRE fx  ";

char * allname =    "laser, stars and fire";
char * cubename =   "all, projected on 3d ";
// char * foo =     "its just a checkpoint";

float modulate, mod_inc;
int count, direction, count2;
int phi1, phi2;

randomize; // in typen.h definiert

signalbehandler(); // Aus meinem OSS Modul: nette Fehlermeldungen, Ende
                   // Eventuell um X-Close und Speicherfreigabe erweitern?
                   // Aber am Ende vom Programm wird sowieso alles freigegeben?
init_x ();  // Kein X-Zugriff wird als "falsche Farbtiefe: -1" gemeldet :-)
oss_open(); // Bei einem Fehler gibts halt keinen Sound, sonst Nix.

/* ------------------- */

  printf( "\n\
 This is FXcube, a tiny Linux demo where I collected my favourite\n\
 oldskool effects on a single 3d spinning cube! Have fun. To skip\n\
 scenes, press keys. Some 100 MHz needed. You need X11 with 16bit\n\
 colors and 1.5 MB of free RAM, plus /dev/dsp for fake music :-).\n\n\
 Copyright 1994-2003 by Eric Auer <eric@coli.uni-sb.de>. This is\n\
 free software - read license.txt or the GPL 2 at www.gnu.org for\
 more!\n\n" );

/* ------------------- */

  testmuster (1, W , H , buffer , 1.0 ); 
  /* alles ausser dem letzten Wert wird nur hier gelesen! */
  init_fps ();

  count = MUSTERCOUNT;
  while ( event_x () && ((count--) > 0) )
  { 
#ifdef USEDSP  
    testmuster (0, 0 , 0 , NULL , dsp2derive(0 /* stdin */) );
#else
    testmuster (0, 0 , 0 , NULL , 1.0 ); /* eh konstant, aber wg fps ...  */
#endif
    next_fps ();
    UPDATE
  };
  
  show_fps (mustername);
  testmuster (-1, 0 , 0 , NULL , 1.0 );

/* ------------------- */

#ifdef FORK
 fprintf(stderr,"OSS Sound wird nun im Fork-Modus aktiviert!\n");
 (void)gettime(); // Zeitbasis starten
 oss_forkplayer(&musikfueller); 
  // -- riskant, wenn 1 Prozess steht,
  // geht die ganze Synchronisation floeten!
#endif
  // alternativ: oss_fillbuffer(void(*player)()), 
  //             oss_send(void * buffer, int length)

/* ------------------- */

  inter_step ( 1 , minimum(W,192) , minimum(H,192) , \
    &GFX( (W-minimum(W,192)) >> 1 , (H-minimum(H,192)) >> 1 , buffer) , 1.0 ); 
  /*  nur hier werden die meisten Werte verwendet  */
  /* Wegen Geschwindigkeit auf 192 x 192 begrenzen */
  init_fps ();

  count = INTERCOUNT;
  while ( event_x () && ((count--) > 0) )
  { inter_step ( 0 , 0 , 0 , NULL , 1.0 );  // Bonuswert ist Zeit-Zoom
    next_fps (); 
    UPDATE 
  }

  show_fps (intername);
  inter_step( -1 , W , H , buffer , 0.0 );

/* ------------------- */

  laser_step ( 1 , minimum( W , ( H << 2 ) / 3 ) , H ,
    &GFX( ( W - minimum( W , ( H << 2 ) / 3 ) ) >> 1 ,
      0 , buffer ) );
    /* nur hier werden die Werte verwendet */
    /* falls mehr als 4/3 H, begrenze aspect ratio */
  init_fps ();

  count = LASERCOUNT;
  while ( event_x () && ((count--) > 0) )
  { laser_step ( 0 , 0 , 0 , NULL ); 
    next_fps (); 
    UPDATE 
  }

  show_fps (lasername);
  laser_step( -1 , 0 , 0 , NULL );

/* ------------------- */

  plasma_step ( 1 , W , H ,
     (W<H) ? &GFX(0, (H-W)>>1, buffer) : &GFX((W-H)>>1, 0, buffer) ,
     240 );
     /* nur hier werden alle Werte verwendet */
     /* letzter Wert ist HIER eine Prozentangabe fuer den Zufall */
     /* tatsaechliche Groesse kann kleiner werden: Ist immer ein */
     /* Quadrat mit Seitenlaenge 1<<n, liegt am Design...        */
  init_fps ();

  count = 0;
  while ( event_x () && (count < PLASMACOUNT) )
  { 
#ifdef BENCHMARK_PLASMASETUP
    plasma_step ( 1 , W , H , buffer , 240 ); 
#endif
    plasma_step ( 0 , 0 , 0 , NULL , count << 2 ); // letzter Wert ist Phase
    count++;
    count &= 255;
    next_fps (); 
    UPDATE 
  }

  show_fps (plasmaname);
  plasma_step( -1 , W , H , buffer , 0 );

/* ------------------- */

  apfel_step ( 1 , minimum(W,224) , minimum(H,224) , \
    &GFX( (W-minimum(W,224)) >> 1 , (H-minimum(H,224)) >> 1 , buffer) , \
    ZIEL1 );  /* nur hier werden alle Werte verwendet - plus Zielwahl  */
    /* auf 224x224 begrenzen, damit es nicht zu lahm wird */
  init_fps ();

  count = INVSTEPS;
  direction = 1;

  phi1 = APFELCOUNT;
  while ( event_x () && ((phi1--) > 0) )
  { modulate = POWi( ZOOM * ZOOM , count ); // jeweils doppelte Schritte
    apfel_step ( 0 , 0 , 0 , NULL , modulate );
    // Bonuswert ist hier der Radius des Bildausschnittes
    if (direction>0) { count++; if (count>=   (STEPS>>1)) direction = -1; }
    else             { count--; if (count<=(INVSTEPS>>1)) direction =  1; };
    next_fps ();
    UPDATE 
  }

  show_fps (apfelname);
  apfel_step( -1 , 0 , 0 , NULL , 0.0 );

/* ------------------- */

  fire_step ( 1 , minimum(W,140) , minimum(H,160) , \
    &GFX( (W-minimum(W,140)) >> 1 , (H-minimum(H,160)) >> 1 , buffer)); 
    /* nur hier werden die Werte verwendet */
    /* auf 140x160 begrenzen, damit es nicht zu lahm wird */
  init_fps ();

  count = FIRECOUNT;
  while ( event_x () && ((count--) > 0) )
  { fire_step ( 0 , 128 , 160 , buffer ); 
    next_fps (); 
    UPDATE 
  }

  show_fps (firename);
  fire_step( -1 , 0 , 0 , NULL );

/* ------------------- */

  star_step ( 1 , W , H , buffer ); /* nur hier werden die Werte verwendet */
  init_fps ();

  count = STARCOUNT;
  while ( event_x () && ((count--) > 0) )
  { star_step ( 0 , W , H , buffer ); 
    next_fps (); 
    UPDATE 
  }

  show_fps (starname);
  star_step( -1 , W , H , buffer );

/* ------------------- */
  
  star_step  ( 1 , W>>1 , H>>1 , &GFX(0        , 0    , buffer));
  laser_step ( 1 , W>>2 , H>>1 , &GFX(    W>>1 , 0    , buffer));
  plasma_step( 1 , W>>2 , H>>1 , &GFX((3*W)>>2 , 0    , buffer) , 240 );

  fire_step  ( 1 , W>>2 , H>>1 , &GFX(0        , H>>1 , buffer));
  testmuster ( 1 , W>>2 , H>>1 , &GFX(    W>>2 , H>>1 , buffer) , 1.0 );
  inter_step ( 1 , W>>2 , H>>1 , &GFX(    W>>1 , H>>1 , buffer) , 1.0 );
  apfel_step ( 1 , W>>2 , H>>1 , &GFX((3*W)>>2 , H>>1 , buffer) , ZIEL2 );

  init_fps ();
  mod_inc = PI2 / 70; /* Zyklusdauer 70 Frames */   // PI2 = 2*PI, math.h
  modulate = 0.0;
  count = INVSTEPS;
  direction = 1;
  count2 = 0;
  phi1 = ALLCOUNT;
  
  // leider sind alle Effekte nicht mehrfach aufrufbar!
  // fire, apfel, inter, testmuster und plasma koennten aber 
  // halbwegs einfach (nicht allzuviele static Variablen)
  // eine Verdoppelungschnittstelle bekommen ???
  while ( event_x () && ((phi1--) > 0) )
  { testmuster ( 0 , 0 , 0 , NULL , 1.42 * (1.33 + sin(2.0 * modulate)) );
    star_step  ( 0 , 0 , 0 , NULL );
    laser_step ( 0 , 0 , 0 , NULL );
    fire_step  ( 0 , 0 , 0 , NULL );
    apfel_step ( 0 , 0 , 0 , NULL , POWi( ZOOM , count ) );
    plasma_step( 0 , 0 , 0 , NULL , count2 );
    inter_step ( 0 , 0 , 0 , NULL , 1.0 * direction );
    modulate += mod_inc;
    if (modulate > PI2) { modulate -= PI2; };
    if (direction>0) { count++; if (count>=   STEPS) direction = -1; }
    else             { count--; if (count<=INVSTEPS) direction =  1; };
    count2 += 7;
    count2 &= 255;
    next_fps ();
    UPDATE
  };

  show_fps (allname);
  
  fire_step  ( -1 , 0 , 0 , NULL );
  star_step  ( -1 , 0 , 0 , NULL );
  laser_step ( -1 , 0 , 0 , NULL );
  testmuster ( -1 , 0 , 0 , NULL , 1.0 );
  apfel_step ( -1 , 0 , 0 , NULL , 0.0 );
  plasma_step( -1 , 0 , 0 , NULL , 0 );
  inter_step ( -1 , 0 , 0 , NULL , 0.0 );
    
/* ------------------- */

  init_warper();

  if (warpinbuf == NULL) {
    oss_close ();
    close_x ();
    printf("Not enough memory for WARPED effects, sorry.\n");
    exit(1);
  }
  
  star_step  ( 1 , W>>1 , H>>1 , &GFX(0        , 0    , warpinbuf));
  laser_step ( 1 , W>>2 , H>>1 , &GFX(    W>>1 , 0    , warpinbuf));
  plasma_step( 1 , W>>2 , H>>1 , &GFX((3*W)>>2 , 0    , warpinbuf) , 240 );

  fire_step  ( 1 , W>>2 , H>>1 , &GFX(0        , H>>1 , warpinbuf));
  testmuster ( 1 , W>>2 , H>>1 , &GFX(    W>>2 , H>>1 , warpinbuf) , 1.0 );
  inter_step ( 1 , W>>2 , H>>1 , &GFX(    W>>1 , H>>1 , warpinbuf) , 1.0 );
  apfel_step ( 1 , W>>2 , H>>1 , &GFX((3*W)>>2 , H>>1 , warpinbuf) , ZIEL2 );

  init_fps ();
  mod_inc = PI2 / 70; /* Zyklusdauer 70 Frames */   // PI2 = 2*PI, math.h
  modulate = 0.0;
  count = INVSTEPS;
  direction = 1;
  count2 = 0;
  phi1 = 0;
  phi2 = 0;
  
  while ( event_x () )
  { testmuster ( 0 , 0 , 0 , NULL , 1.42 * (1.33 + sin(2.0 * modulate)) );
    star_step  ( 0 , 0 , 0 , NULL );
    laser_step ( 0 , 0 , 0 , NULL );
    fire_step  ( 0 , 0 , 0 , NULL );
    apfel_step ( 0 , 0 , 0 , NULL , POWi( ZOOM , count ) );
    plasma_step( 0 , 0 , 0 , NULL , count2 );
    inter_step ( 0 , 0 , 0 , NULL , 1.0 * direction );
    modulate += mod_inc;
    if (modulate > PI2) { modulate -= PI2; };
    if (direction>0) { count++; if (count>=   STEPS) direction = -1; }
    else             { count--; if (count<=INVSTEPS) direction =  1; };
    count2 += 7;
    count2 &= 255;
    next_fps ();
    
      /* argument: 2 cube rotation integers */
    do_warper(phi1, phi2);
    phi1 += 3;
    phi2 += 7;
    if (phi1 >= 360) phi1 -= 360;
    if (phi2 >= 360) phi2 -= 360;
    if ((phi1 > 356) && (phi2 > 352) && (count2 > 220))
      break; /* break out of the loop when it gets boring */
      /* tune this by modifying the count2 comparison */
      /* example: 200 for 360 iterations, 220 for 2500... */
    
    UPDATE
  };

  show_fps (cubename);

  fire_step  ( -1 , 0 , 0 , NULL );
  star_step  ( -1 , 0 , 0 , NULL );
  laser_step ( -1 , 0 , 0 , NULL );
  testmuster ( -1 , 0 , 0 , NULL , 1.0 );
  apfel_step ( -1 , 0 , 0 , NULL , 0.0 );
  plasma_step( -1 , 0 , 0 , NULL , 0 );
  inter_step ( -1 , 0 , 0 , NULL , 0.0 );
  
  close_warper();
    
/* ------------------- */

  oss_close ();
  close_x ();
  
  exit (0);
  
}
