/*
 * This is FXcube, a tiny Linux demo where I collected my favourite
 * oldskool effects on a single 3d spinning cube! Have fun. To skip
 * scenes, press keys. Some 100 MHz needed. You need X11 with 16bit
 * colors and 1.5 MB of free RAM, plus /dev/dsp for fake music :-).
 * Copyright 1994-2003 by Eric Auer <eric@coli.uni-sb.de>.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should find a copy of the GNU General Public License on the
 * site where you have downloaded this source. If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307 USA
 * (or try http://www.gnu.org/licenses/licenses.html at www.gnu.org).
 */

#include "oss.h"

static int DSP=-1; // Filehandle
static int CHILD=0; // forked child
static int rausda=0; // Signal aufzuhoeren (auch ans Child)

/* ----------------------------------------------------------------- */

void auweia(int sig)
{ 
  if ((sig!=SIGTERM) && (sig!=SIGQUIT) && (sig!=SIGINT))
  {
    switch (sig) {
       case SIGFPE: {
         fprintf(stderr, "Gleitkomma-Ausrutscher! ");
         break;
         }
       case SIGSEGV: {
         fprintf(stderr, "Illegaler Grenzuebertritt! ");
         break;
         }
       case SIGILL: {
         fprintf(stderr, "Befehlsverweigerung! ");
         break;
         }
       case SIGIOT: {
         fprintf(stderr, "Keine Ein-Ausgabe-Begabung! ");
         break;
         }
       case SIGSTKFLT: {
         fprintf(stderr, "Verwirrung beim Stapeln! ");
         break;
         }
    }
    fprintf(stderr, "Signal %i klingelt fuer %i, Kind von %i!\n",\
     sig, getpid(), getppid());
    fprintf(stderr,"TERM=%i QUIT=%i ALRM=%i FPE=%i SEGV=%i ILL=%i INT=%i\
 IOT=%i ABRT=%i STKFLT=%i\n", \
     SIGTERM, SIGQUIT, SIGALRM, SIGFPE, SIGSEGV, SIGILL, SIGINT, SIGIOT, \
     SIGABRT, SIGSTKFLT );
  }
  else
  { if (sig==SIGTERM) fprintf(stderr,"Sigterm (%i) caught. Bye bye!\n", sig);
    if (sig==SIGQUIT) fprintf(stderr,"Sigquit (%i) caught. Bye bye!\n", sig);
    if (sig==SIGINT) fprintf(stderr,"Sigint[^C] (%i) caught. Bye bye!\n", sig);
  };

  rausda++; // also mindestens 1 - zur Info ans Child!

  if (CHILD) 
  { int childstat, whichchild;
    sleep(1);
    whichchild = waitpid(CHILD,&childstat,WNOHANG);
    // sys/types.h und sys/wait.h
    if (whichchild == -1) { perror("waitpid error: "); };
    // 0, -1, <-1 sind Sonder-PIDs hier
    if (whichchild > 0)
       { fprintf(stderr,"Child %i beendet, [%i]\n" , whichchild , childstat); 
       }
    else
       { whichchild = kill(CHILD,0); // Sonder-PID wie oben, 0 ist Sonder-Sig
         if ( (whichchild==-1) && (errno==ESRCH) ) 
            {
              perror("Finde Child nicht mehr?");
            } else {
              if (whichchild==-1)
                perror("Fehler beim Child-Killen"); };
         if (kill(CHILD, SIGKILL)==-1) {
           perror("Konnte Child nicht killen");
         } else {
           if (CHILD != 0) {
             fprintf(stderr, "Error killing child (%i)?\n", CHILD);
           } else {
             fprintf(stderr, "Kind ist fertig.\n");
           }
         };
       };
  };

  if (rausda > 2) {
    fprintf(stderr, "Habe QUIT ueberlebt! 10 9 8 7 6 5 4 3 2 1...\n");
    kill(getpid(), SIGKILL);    /* kill definitely, SIGQUIT failed. */
      /* (added 3/2003, because a task may cause lots of SEGV / IOT */
      /*  before it reads the rausda flag for the next time...)     */
  };

  if (rausda > 1) {
    fprintf(stderr, "Habe abort() verpennt. Jetzt aber echt QUIT.\n");
    kill(getpid(), SIGQUIT);        /* kill harder, abort() failed. */
      /* (added 3/2003, because a task may cause lots of SEGV / IOT */
      /*  before it reads the rausda flag for the next time...)     */
  };

  if ((sig==SIGSEGV) || (sig==SIGILL) || (sig==SIGILL) || (sig==SIGABRT) || \
      (sig==SIGIOT) || (rausda>1))
  { 
    /* fprintf(stderr,"Da muss noch wo ein Bug versteckt sein.\n"); */
    abort();
  };

  rausda++; // mehr als 1 Quit? Dann Not-Aus annehmen...
  if (DSP!=-1) oss_close();

  if ((sig==SIGQUIT) || (sig==SIGTERM) || (sig==SIGINT))
  { printf("Bis bald! [%i, Kind von %i]\n",getpid(),getppid()); _exit(0); 
  }
    // Exit vererbt alle Kinder an Init und der Elternprozess kriegt SIGCHLD
    // _exit laesst dabei atexit-Aufrufe weg, aber schliesst alle Dateien
  else
  { 
    printf("Ein Signal wurde uebersehen, wie peinlich...\n"); _exit(1);
  };
};

void child_ack(int sig)
{
  if (CHILD == getpid()) {
    fprintf(stderr, "Hey! I am waiting for me my own child?\n");
  };
  if (CHILD)
  { fprintf(stderr,"Child (%i) has terminated, %i waiting for it.\n",
      CHILD, getpid());
    if (waitpid(CHILD,NULL,WNOHANG) == CHILD) { CHILD = 0; };
  };
    signal(SIGCHLD, &child_ack);
};

void signalbehandler() // siehe man 7 signal
{ signal(SIGINT, &auweia);
#ifdef HANDLE_ALRM
  signal(SIGALRM, &auweia);
#endif
  signal(SIGQUIT, &auweia);
  signal(SIGTERM, &auweia);
  signal(SIGFPE, &auweia);
  signal(SIGSEGV, &auweia);
  signal(SIGILL, &auweia);
  signal(SIGABRT, &auweia);
  signal(SIGIOT, &auweia);
  signal(SIGSTKFLT, &auweia);
  signal(SIGCHLD, &child_ack);
  signal(SIGCLD, &child_ack);
}; // PS: ..TSTP, TTIN, TTOU sind Stop weil bg und TTY-Zugriff...
   // SIGSTOP und SIGKILL koennen nicht abgefangen werden... SIGCONT = ?

/* ----------------------------------------------------------------- */

static int timebase=-1;

int gettime()
{ struct timeval tod;
  int mytime;
  gettimeofday( &tod , NULL );
  mytime = (tod.tv_sec*HZ) + (tod.tv_usec*(HZ/1000000));
  if (timebase<0) timebase = mytime;
  return mytime-timebase;
};

/* ----------------------------------------------------------------- */

void oss_close()
{ if (CHILD) 
     { (void)kill(CHILD , SIGTERM); sleep(1); (void)kill(CHILD, SIGTERM); };
  if (DSP!=-1)
  { ioctl(DSP , SNDCTL_DSP_RESET , 0); close(DSP); };
};

void oss_send(void * buffer, int length)
{ int what, why;
  if (DSP!=-1) 
  { do 
    {  what = write(DSP , buffer , length ); 
       if (what==-1) { why=errno;
                       perror("DSP write error "); 
                     }
       else { if (what!=length) 
                 { fprintf(stderr,"Only %i of %i bytes written to DSP",\
                                   what,length);
                 };
              why = 0;
            };
    } while ((why==EINTR) || (why==EAGAIN)); // EAGAIN - bei O_NONBLOCK ...
      // EINTR: Eintreffendes Signal hat den ganzen write zunichte gemacht!
  };
}; /* native format: bytes */

// int oss_gettime() // zur Zeit toter Code
// { char started=0;  count_info ci;  int btmr=0;  static int lastby=0;
//   for(;;) { ioctl(DSP , SNDCTL_DSP_GETOPTR , &ci);
//             if ( (!started) && (ci.bytes==65536) ) player(); };
//   started=1;  return (ci.bytesr*HZ) / (SampRate*ByPerSamp); };

void oss_fillbuffer(void(*player)())
{ audio_buf_info a;
  if ( (DSP==-1) || (player==NULL) ) return;
  for(;;)
  { ioctl(DSP , SNDCTL_DSP_GETOSPACE , &a);
    if (a.fragments < a.fragstotal/2) break;
    player();
  };
};

void oss_forkplayer(void(*player)())
{ register int whoami;
  if ( (!player) || (DSP==-1) || (CHILD)) return;
  // player(); - gefaehrlich! Parent startet auch den Player, wozu?
  whoami = fork();
  if (whoami!=-1) // fork successful?
  { if (!whoami) // am I the child?
    { 
#ifdef HANDLE_ALRM
      signal( SIGALRM , SIG_IGN );
#endif
      for(;;) 
      { player();
        if (rausda) exit(0);
        if (getppid()==1) 
        { fprintf(stderr,"Wurde von init adoptiert - Hauptprozess weg?\n");
          fprintf(stderr,"Dieser Prozess (%i)wird jetzt auch beendet\n",\
          getpid()); oss_close(); exit(1);
        };
      };
    }
    else // or am I the parent process?
    { CHILD = whoami;
      fprintf(stderr,"Successfully spawned player child. PID:  %i\n", whoami);
    };
  }
  else 
  { perror("Sound player could not fork ");  /* -> sys_errlist[errno]  */ 
  };
};

void oss_open()
{ int tmp = (NUMFRAGS<<16) | FRAGSIZE;
  int bits = ByPerSamp << 3;
  int channels = StereoChn;
  int rate = SampRate;
  char dspname[] = OSS_FILE;
  DSP = -1;
  if ( ( DSP = open( OSS_FILE , O_WRONLY|O_SYNC /*O_NDELAY*/ , 0 ) ) < 0 ) 
     // lieber SYNC als NONBLOCK/NDELAY, damit alles im Takt bleibt!
     // evtl. mit Timeout "Steckenbleiben" abfangen...
  { fprintf(stderr, "Kann %s ", dspname);
    perror("nicht oeffnen!"); 
    DSP = -1; return; };
    // (perror decodiert gleich die interne errno zu Text)
    // alle folgenden Fehler sind nicht so heftig - wenn
    // z.B. der Soundtreiber eine andere Sorte oder sogar
    // eine Pipe sein sollte, trotzdem nicht aufgeben...
  fprintf(stderr, "Sounddevice als Handle Nr. %i geoeffnet\n", DSP);
  
  if( ioctl( DSP , SNDCTL_DSP_SETFRAGMENT , &tmp      ) < 0 )
  { fprintf(stderr,"setfragment?\n"); };
  if( ioctl( DSP , SNDCTL_DSP_SAMPLESIZE  , &bits     ) < 0 )
  { fprintf(stderr,"samplesize?\n"); };
  if( ioctl( DSP , SNDCTL_DSP_STEREO      , &channels ) < 0 )
  { fprintf(stderr,"dspstereo?\n"); };
  if( ioctl( DSP , SNDCTL_DSP_SPEED       , &rate     ) < 0 )
  { fprintf(stderr,"dspspeed?\n"); };
};
